import math
import torch
import datetime
import numpy as np
from torch import nn
from torch.nn import Module, Parameter
import torch.nn.functional as F


class OngoingSessionGraph(Module):
    def __init__(self, hidden_size, step=1, dp3=0.1):
        super(OngoingSessionGraph, self).__init__()
        self.step = step
        self.dp3 = dp3
        self.hidden_size = hidden_size
        self.input_size = hidden_size * 2
        self.gate_size = 3 * hidden_size

        self.w_ih = Parameter(torch.Tensor(self.gate_size, self.input_size))
        self.w_hh = Parameter(torch.Tensor(self.gate_size, self.hidden_size))
        self.b_ih = Parameter(torch.Tensor(self.gate_size))
        self.b_hh = Parameter(torch.Tensor(self.gate_size))
        self.b_iah = Parameter(torch.Tensor(self.hidden_size))
        self.b_oah = Parameter(torch.Tensor(self.hidden_size))

        self.linear_edge_in = nn.Linear(self.hidden_size, self.hidden_size, bias=True)
        self.linear_edge_out = nn.Linear(self.hidden_size, self.hidden_size, bias=True)
        self.linear_edge_f = nn.Linear(self.hidden_size, self.hidden_size, bias=True)
        self.highway = nn.Linear(2 * self.hidden_size, self.hidden_size, bias=True)
        self.sigmoid = nn.Sigmoid()
        self.dropout = nn.Dropout(self.dp3)

    def GNNCell(self, A, hidden):
        input_in = torch.matmul(A[:, :, :A.shape[1]], self.linear_edge_in(hidden)) + self.b_iah
        input_out = torch.matmul(A[:, :, A.shape[1]: 2 * A.shape[1]], self.linear_edge_out(hidden)) + self.b_oah
        inputs = torch.cat([input_in, input_out], 2)
        gi = F.linear(inputs, self.w_ih, self.b_ih)
        gh = F.linear(hidden, self.w_hh, self.b_hh)
        i_r, i_i, i_n = gi.chunk(3, 2)
        h_r, h_i, h_n = gh.chunk(3, 2)
        resetgate = torch.sigmoid(i_r + h_r)
        inputgate = torch.sigmoid(i_i + h_i)
        newgate = torch.tanh(i_n + resetgate * h_n)

        hy = newgate + inputgate * (hidden - newgate)
        return hy

    def forward(self, A, hidden2):
        origin_hidden = hidden2
        hidden = hidden2
        for i in range(self.step):
            hidden = self.GNNCell(A, hidden)
            hidden = self.dropout(hidden)
        final_hidden = torch.cat((origin_hidden, hidden), dim=-1)
        g = self.sigmoid(self.highway(final_hidden))
        hidden = g * origin_hidden + (1 - g) * hidden
        return hidden
