import numpy as np


def data_masks(all_usr_pois, item_tail):
    us_pois, us_msks, node_num, length = [], [], [], []
    us_lens = [len(upois) for upois in all_usr_pois]
    len_max = max(us_lens)
    for upois, le in zip(all_usr_pois, us_lens):
        us_pois.append(upois + item_tail * (len_max - le))
        us_msks.append([1] * le + [0] * (len_max - le))
    node_num = [len(np.unique(upois)) for upois in us_pois]
    return us_pois, us_msks, len_max, node_num, us_lens


def split_validation(train_set, valid_portion):
    train_set_x, train_set_y = train_set
    n_samples = len(train_set_x)
    sidx = np.arange(n_samples, dtype='int32')
    np.random.shuffle(sidx)
    n_train = int(np.round(n_samples * (1. - valid_portion)))
    valid_set_x = [train_set_x[s] for s in sidx[n_train:]]
    valid_set_y = [train_set_y[s] for s in sidx[n_train:]]
    train_set_x = [train_set_x[s] for s in sidx[:n_train]]
    train_set_y = [train_set_y[s] for s in sidx[:n_train]]

    return (train_set_x, train_set_y), (valid_set_x, valid_set_y)


class Data():
    def __init__(self, data, opt, shuffle=False):
        cut_inputs = [input_[-opt.recent:] for input_ in data[0]]
        inputs, mask, len_max, node_num, sess_len = data_masks(cut_inputs, [0])
        self.node_num = np.asarray(node_num)
        self.sess_len = np.asarray(sess_len)
        self.inputs = np.asarray(inputs)
        self.mask = np.asarray(mask)
        self.len_max = len_max
        self.targets = np.asarray(data[1])
        self.length = len(inputs)
        self.shuffle = shuffle
        self.aug = opt.aug

    def generate_batch(self, batch_size):
        if self.shuffle:
            shuffled_arg = np.arange(self.length)
            np.random.shuffle(shuffled_arg)
            self.sess_len = self.sess_len[shuffled_arg]
            self.node_num = self.node_num[shuffled_arg]
            self.inputs = self.inputs[shuffled_arg]
            self.mask = self.mask[shuffled_arg]
            self.targets = self.targets[shuffled_arg]
        n_batch = int(self.length / batch_size)
        if self.length % batch_size != 0:
            n_batch += 1
        slices = np.split(np.arange(n_batch * batch_size), n_batch)
        slices[-1] = slices[-1][:(self.length - batch_size * (n_batch - 1))]
        return slices

    def get_slice_gnn(self, i):
        inputs, mask, targets, srgnn_node_num, session_len = self.inputs[i], self.mask[i], self.targets[i], \
                                                             self.node_num[i], self.sess_len[i]
        items, n_node, A, alias_inputs = [], [], [], []
        max_n_node = np.max(srgnn_node_num)
        all_items = list(np.unique(inputs))
        if 0 in all_items:
            all_items.remove(0)
        all_items.append(0)
        all_num_node = len(all_items)
        all_items = np.array(all_items)
        adj = np.zeros((all_num_node, all_num_node))
        adj2 = np.zeros((all_num_node, all_num_node))
        sess_has_item = [[all_num_node - 1] * self.len_max]
        item_belong_sess = [set() for _ in range(all_num_node)]
        cnt = 1

        for u_input in inputs:
            node = np.unique(u_input)
            items.append(node.tolist() + (max_n_node - len(node)) * [0])
            u_A = np.zeros((max_n_node, max_n_node))
            cur_alias = []
            cur_sess_has = []
            for i in np.arange(len(u_input) - 1):
                p = np.where(all_items == u_input[i])[0][0]
                u = np.where(node == u_input[i])[0][0]
                cur_sess_has.append(p)
                item_belong_sess[p].add(cnt)
                cur_alias.append(u)
                if u_input[i + 1] == 0:
                    break
                q = np.where(all_items == u_input[i + 1])[0][0]
                adj[p][q] = 1
                v = np.where(node == u_input[i + 1])[0][0]
                u_A[u][v] = 1
            if u_input[-1] != 0:
                p = np.where(all_items == u_input[-1])[0][0]
                item_belong_sess[p].add(cnt)
                cur_sess_has.append(np.where(all_items == u_input[-1])[0][0])
                cur_alias.append(np.where(node == u_input[-1])[0][0])
            cur_sess_has += [all_num_node - 1] * (self.len_max - len(cur_sess_has))
            cur_alias += [0] * (len(u_input) - len(cur_alias))
            alias_inputs.append(cur_alias)
            sess_has_item.append(cur_sess_has)

            u_sum_in = np.sum(u_A, 0)
            u_sum_in[np.where(u_sum_in == 0)] = 1
            u_A_in = np.divide(u_A, u_sum_in)
            u_sum_out = np.sum(u_A, 1)
            u_sum_out[np.where(u_sum_out == 0)] = 1
            u_A_out = np.divide(u_A.transpose(), u_sum_out)
            u_A = np.concatenate([u_A_in, u_A_out]).transpose()
            A.append(u_A)

            cnt += 1

        l = 0
        for i in range(len(item_belong_sess)):
            item_belong_sess[i] = list(set(item_belong_sess[i]))
            l = max(l, len(item_belong_sess[i]))
        for i in item_belong_sess:
            i += [0] * (l - len(i))

        u_sum_in = np.sum(adj, 0)
        u_sum_in[np.where(u_sum_in == 0)] = 1
        u_A_in = np.divide(adj, u_sum_in)
        u_sum_out = np.sum(adj, 1)
        u_sum_out[np.where(u_sum_out == 0)] = 1
        u_A_out = np.divide(adj.transpose(), u_sum_out)
        adjacency = np.concatenate([u_A_in, u_A_out]).transpose()

        adj2[:, :] = adj[:, :]
        a = np.random.random((adj2.shape[0], adj2.shape[1]))
        a[a >= self.aug] = 1
        a[a < self.aug] = 0
        adj2 *= a
        u_sum_in = np.sum(adj2, 0)
        u_sum_in[np.where(u_sum_in == 0)] = 1
        u_A_in = np.divide(adj2, u_sum_in)
        u_sum_out = np.sum(adj2, 1)
        u_sum_out[np.where(u_sum_out == 0)] = 1
        u_A_out = np.divide(adj2.transpose(), u_sum_out)
        adjacency2 = np.concatenate([u_A_in, u_A_out]).transpose()

        return alias_inputs, A, adjacency, adjacency2, mask, items, targets, sess_has_item, item_belong_sess, all_items, session_len
